#define DRIVER_VERSION  "v1.1.0"
#define DRIVER_AUTHOR   "John Klug <john.klug@multitech.com>"
#define DRIVER_DESC "MTS General Purpose I/O Accessory Card"
#define DRIVER_NAME "mtac-gpiob"

#include <linux/types.h>
#include <linux/gpio.h>
#include <linux/platform_device.h>
#include <linux/kmod.h>
#include <linux/bitops.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/spi/spi.h>
#include <linux/mtac.h>
#include <linux/mts_io.h>

static struct gpio_pin gpio_pins_mtcdt_mtac_gpiob[] = {
	// gpio pins for Accessory Card 1
	{
		.name = "AP1_RESET",
		.pin = {
			.gpio = AT91_PIN_PB12,
			.flags = GPIOF_OUT_INIT_HIGH,
			.label = "ap1-reset",
		}
	},
	{
		.name = "AP1_GPIO1",
		.pin = {
			.gpio = AT91_PIN_PC6,
			.flags = GPIOF_OUT_INIT_LOW,
			.label = "ap1-gpio1",
		},
		.active_low = 1,
	},
	{
		.name = "AP1_GPIO2",
		.pin = {
			.gpio = AT91_PIN_PC7,
			.flags = GPIOF_OUT_INIT_LOW,
			.label = "ap1-gpio2",
		}
	},
	{
		.name = "AP1_GPIO3",
		.pin = {
			.gpio = AT91_PIN_PC8,
			.flags = GPIOF_OUT_INIT_LOW,
			.label = "ap1-gpio3",
		}
	},
	{
		.name = "AP1_GPIO4",
		.pin = {
			.gpio = AT91_PIN_PC9,
			.flags = GPIOF_OUT_INIT_LOW,
			.label = "ap1-gpio4",
		}
	},
	{
		.name = "AP1_INTERRUPT1",
		.pin = {
			.gpio = AT91_PIN_PB14,
			.flags = GPIOF_IN,
			.label = "ap1-interrupt1",
		}
	},
	{
		.name = "AP1_INTERRUPT2",
		.pin = {
			.gpio = AT91_PIN_PB15,
			.flags = GPIOF_IN,
			.label = "ap1-interrupt2",
		}
	},
	{
		.name = "AP1_GPS_PPS",
		.pin = {
			.gpio = AT91_PIN_PA29,
			.flags = GPIOF_IN,
			.label = "ap1-gps-pps",
		}
	},

	// gpio pins for Accessory Card 2
	{
		.name = "AP2_RESET",
		.pin = {
			.gpio = AT91_PIN_PB13,
			.flags = GPIOF_OUT_INIT_HIGH,
			.label = "ap2-reset",
		}
	},
	{
		.name = "AP2_GPIO1",
		.pin = {
			.gpio = AT91_PIN_PC20,
			.flags = GPIOF_OUT_INIT_LOW,
			.label = "ap2-gpio1",
		},
		.active_low = 1,
	},
	{
		.name = "AP2_GPIO2",
		.pin = {
			.gpio = AT91_PIN_PC21,
			.flags = GPIOF_OUT_INIT_LOW,
			.label = "ap2-gpio2",
		}
	},
	{
		.name = "AP2_GPIO3",
		.pin = {
			.gpio = AT91_PIN_PC22,
			.flags = GPIOF_OUT_INIT_LOW,
			.label = "ap2-gpio3",
		}
	},
	{
		.name = "AP2_GPIO4",
		.pin = {
			.gpio = AT91_PIN_PC23,
			.flags = GPIOF_OUT_INIT_LOW,
			.label = "ap2-gpio4",
		}
	},
	{
		.name = "AP2_INTERRUPT1",
		.pin = {
			.gpio = AT91_PIN_PB17,
			.flags = GPIOF_IN,
			.label = "ap2-interrupt1",
		}
	},
	{
		.name = "AP2_INTERRUPT2",
		.pin = {
			.gpio = AT91_PIN_PB18,
			.flags = GPIOF_IN,
			.label = "ap2-interrupt2",
		}
	},
	{
		.name = "AP2_GPS_PPS",
		.pin = {
			.gpio = AT91_PIN_PA6,
			.flags = GPIOF_IN,
			.label = "ap2-gps-pps",
		}
	},

	{ },
};

enum spi_devices {
	din = 0,
	dout = 1,
	adc = 2,
};

struct spi_device *gpiob_spi[NUM_AP][3];
struct spi_driver gpiob_spi_drivers[NUM_AP][3];

static u8 spi_ap_dout_value[NUM_AP];
static DEFINE_MUTEX(spi_ap_dout_mutex);
static unsigned int ap_dout_max_speed_hz = 1 * 1000 * 1000;
module_param(ap_dout_max_speed_hz, uint, S_IRUGO);
MODULE_PARM_DESC(
	ap_dout_max_speed_hz,
	"Maximum clock rate to be used with this device (default: 1 MHz)"
);

static unsigned int ap_din_max_speed_hz = 1 * 1000 * 1000;
module_param(ap_din_max_speed_hz, uint, S_IRUGO);
MODULE_PARM_DESC(
	ap_din_max_speed_hz,
	"Maximum clock rate to be used with this device (default: 1 MHz)"
);

static unsigned int ap_adc_max_speed_hz = 20 * 1000 * 1000;
module_param(ap_adc_max_speed_hz, uint, S_IRUGO);
MODULE_PARM_DESC(
	ap_adc_max_speed_hz,
	"Maximum clock rate to be used with this device (default: 20 MHz)"
);

static bool gpiob_get_dev_info_from_modalias(const char* modalias, int* port, char* buf) {
	sscanf(modalias, "mts-io-ap%d-%s", port, buf);

	return true;
}

/* Generic SPI functions */
static inline int spi_writen(struct spi_device *spi, const u8 *buf, size_t len)
{
	int tmp;
	u8 *tx;

	tx = kmalloc(len, GFP_KERNEL);
	if (!tx) {
		return -ENOMEM;
	}

	memcpy(tx, buf, len);
	tmp = spi_write(spi, tx, len);

	kfree(tx);

	return tmp;
}

static inline int spi_readn(struct spi_device *spi, u8 *buf, size_t len)
{
	int tmp;
	u8 *rx;

	rx = kmalloc(len, GFP_KERNEL);
	if (!rx) {
		return -ENOMEM;
	}

	tmp = spi_read(spi, rx, len);
	memcpy(buf, rx, len);

	kfree(rx);

	return tmp;
}

static int mts_spi_ap_probe(struct spi_device *spi)
{
	int tmp;
	int port;
	int port_index;
	char buf[16];
	enum spi_devices dev;

	gpiob_get_dev_info_from_modalias(spi->modalias, &port, buf);
	port_index = port - 1;
	if (port < 1 || port > NUM_AP) {
		log_error("port %d is invalid", port);
		return -ENODEV;
	}

	if (strstr(buf, "dout")) {
		dev = dout;
		spi->max_speed_hz = ap_dout_max_speed_hz;
		spi->mode = 0;
	} else if (strstr(buf, "din")) {
		dev = din;
		spi->max_speed_hz = ap_din_max_speed_hz;
		spi->mode = SPI_CPOL;
	} else if (strstr(buf, "adc")) {
		dev = adc;
		spi->max_speed_hz = ap_adc_max_speed_hz;
		spi->mode = 0;
	} else {
		log_error("unknown gpiob spi device type [%s]", buf);
		return -ENODEV;
	}

	gpiob_spi[port_index][dev] = spi;

	tmp = spi_setup(gpiob_spi[port_index][dev]);
	if (tmp < 0) {
		log_error("spi_setup ap %d [%s] failed", port, buf);
		return tmp;
	}

	if (dev == dout) {
		spi_ap_dout_value[port_index] = 0x00;
		spi_writen(gpiob_spi[port_index][dev], &spi_ap_dout_value[port_index], 1);
	}

	return 0;
}

static int mts_spi_ap_remove(struct spi_device *spi)
{
	int port;
	int port_index;
	char buf[16];

	gpiob_get_dev_info_from_modalias(spi->modalias, &port, buf);
	port_index = port - 1;
	if (port < 1 || port > NUM_AP) {
		log_error("port %d is invalid", port);
		return -ENODEV;
	}

	if (strstr(buf, "dout")) {
		gpiob_spi[port_index][dout] = NULL;
	} else if (strstr(buf, "din")) {
		gpiob_spi[port_index][din] = NULL;
	} else if (strstr(buf, "adc")) {
		gpiob_spi[port_index][adc] = NULL;
	} else {
		log_error("unknown gpiob spi device type [%s]", buf);
		return -ENODEV;
	}

	return 0;
}

static char* gpiob_gpio_pin_name_by_attr_name(const char* name, int port) {
	switch (port) {
		case port_1:
			if (! strcmp(name, "led1")) {
				return "ap1-gpio3";
			} else if (! strcmp(name, "led2")) {
				return "ap1-gpio4";
			} else if (! strcmp(name, "dout-enable")) {
				return "ap1-gpio1";
			} else if (! strcmp(name, "reset")) {
				return "ap1-reset";
			} else {
				log_error("attribute name [%s] is invalid for GPIOB in port %d", name, port);
				return "";
			}

		case port_2:
			if (! strcmp(name, "led1")) {
				return "ap2-gpio3";
			} else if (! strcmp(name, "led2")) {
				return "ap2-gpio4";
			} else if (! strcmp(name, "dout-enable")) {
				return "ap2-gpio1";
			} else if (! strcmp(name, "reset")) {
				return "ap2-reset";
			} else {
				log_error("attribute name [%s] is invalid for GPIOB in port %d", name, port);
				return "";
			}
	}
	log_error("gpiob: Invalid port number");
	return "";
}

static ssize_t mts_attr_show_ap_din(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
	int tmp;
	int channel;
	int port;
	int port_index;
	u8 bit;
	u8 byte;

	sscanf(attr->attr.name, "din%d", &channel);
	if (channel < 0 || channel > 3) {
		log_error("channel %d is invalid", channel);
		return -ENOENT;
	}

	port = mtac_port_from_kobject(kobj);
	if (port < 0) {
		log_error("mtac_port_from_kobject returned %d", port);
		return -EINVAL;
	}
	port_index = port - 1;

	bit = BIT(channel);

	tmp = spi_readn(gpiob_spi[port_index][din], &byte, 1);
	if (tmp) {
		log_error("spi_read failed %d", tmp);
		return tmp;
	}

	tmp = byte & bit ? 1 : 0;

	return sprintf(buf, "%d\n", tmp);
}

static ssize_t mts_attr_store_ap_dout(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t count)
{
	int value;
	int channel;
	int port;
	int port_index;
	u8 bit;

	sscanf(attr->attr.name, "dout%d", &channel);
	if (channel < 0 || channel > 3) {
		log_error("channel %d is invalid", channel);
		return -ENOENT;
	}

	port = mtac_port_from_kobject(kobj);
	if (port < 0) {
		log_error("mtac_port_from_kobject returned %d", port);
		return -EINVAL;
	}
	port_index = port - 1;

	bit = BIT(channel);

	if (sscanf(buf, "%i", &value) != 1) {
		log_error("accessory card dout attr invalid argument %d", value);
		return -EINVAL;
	}

	mutex_lock(&spi_ap_dout_mutex);

	if (value) {
		spi_ap_dout_value[port_index] &= ~bit;
	} else {
		spi_ap_dout_value[port_index] |= bit;
	}

	spi_writen(gpiob_spi[port_index][dout], &spi_ap_dout_value[port_index], 1);

	mutex_unlock(&spi_ap_dout_mutex);

	return count;
}

static ssize_t mts_attr_show_ap_dout(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
	int value;
	int channel;
	int port;
	int port_index;
	u8 bit;

	sscanf(attr->attr.name, "dout%d", &channel);
	if (channel < 0 || channel > 3) {
		log_error("channel %d is invalid", channel);
		return -ENOENT;
	}

	port = mtac_port_from_kobject(kobj);
	if (port < 0) {
		log_error("mtac_port_from_kobject returned %d", port);
		return -EINVAL;
	}
	port_index = port - 1;

	bit = BIT(channel);

	mutex_lock(&spi_ap_dout_mutex);

	value = spi_ap_dout_value[port_index] & bit ? 0 : 1;

	mutex_unlock(&spi_ap_dout_mutex);

	return sprintf(buf, "%d\n", value);
}

static ssize_t mts_attr_show_ap_adc(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
	int tmp;
	int tx_data;
	int rx_data;
	int channel;
	int port;
	int port_index;
	int channel_mask = 0x0180;  /* 0b 0000 0001 1000 0000 */
	int manual_mode = 0x1840;   /* 0b 0001 1000 0100 0000 */
	uint8_t tx[2];
	uint8_t rx[2];

	memset(tx, 0, sizeof(tx));
	memset(rx, 0, sizeof(rx));

	sscanf(attr->attr.name, "adc%d", &channel);
	if (channel < 0 || channel > 2) {
		log_error("channel %d is invalid", channel);
		return -ENOENT;
	}

	port = mtac_port_from_kobject(kobj);
	if (port < 0) {
		log_error("mtac_port_from_kobject returned %d", port);
		return -EINVAL;
	}
	port_index = port - 1;

	/* 1st transfer to set up (5V reference, channel to read from) */
	tx_data = manual_mode | ((channel << 7) & channel_mask);
	tx[0] = tx_data >> 8;
	tx[1] = tx_data & 0xFF;
	tmp = spi_writen(gpiob_spi[port_index][adc], tx, 2);
	if (tmp) {
		log_error("spi_write failed %d", tmp);
		return tmp;
	}

	/* 2nd transfer to clock chip for ADC conversion
	 * this can be a throw-away read or an empty write,
	 * the ADC just needs the clock running so it can convert */
	tx[0] = 0;
	tx[1] = 0;
	tmp = spi_writen(gpiob_spi[port_index][adc], tx, 2);
	if (tmp) {
		log_error("2nd spi_write failed %d", tmp);
		return tmp;
	}

	/* 3rd transfer to read data */
	tmp = spi_readn(gpiob_spi[port_index][adc], rx, 2);
	if (tmp) {
		log_error("spi_read failed %d", tmp);
		return tmp;
	}
	rx_data = ((rx[0] & 0x0F) << 8) | (rx[1] & 0xFF);

	return sprintf(buf, "%lu\n", (unsigned long) rx_data);
}

static bool gpiob_spi_driver_setup(struct spi_driver *driver, const char *driver_name) {
	char* name = kstrdup(driver_name, GFP_KERNEL);
	if (! name) {
		log_error("GFP_KERNEL dup failed for driver [%s]", driver_name);
		return false;
	}
	driver->driver.name = name;
	driver->driver.bus = &spi_bus_type;
	driver->driver.owner = THIS_MODULE;
	driver->probe = mts_spi_ap_probe;
	driver->remove = mts_spi_ap_remove;

	return true;
}

// 4 digital inputs
// 4 digital outputs
// 3 analog to digital
// 2 LEDs
// 1 digital out enable
// 1 reset
// 1 vendor-id
// 1 product-id
// 1 device-id
// 1 hw-version
// NULL
static int ap_gpiob_attrs_size = 20;

static bool gpiob_setup(enum ap port) {
	int i;
	int port_index = port - 1;
	int index = 0;
	int count = 0;
	int ret;
	char buf[32];
	struct kobj_attribute* attr;
	struct attribute **attrs;
	struct kobject *subdir;

	log_info("loading GPIOB accessory card in port %d", port);

	sprintf(buf, "ap%d", port);
	subdir = kobject_create_and_add(buf, &mts_io_platform_device->dev.kobj);
	if (! subdir) {
		log_error("kobject_create_and_add for port %d failed", port);
		return false;
	}

	mtac_set_port_pins(port_index,gpio_pins_mtcdt_mtac_gpiob,subdir);

	// create the link to the apX directory this card is in
	// if we're in the first slot, we get plain "gpiob"
	// if we're in a different slot, we might need to use "gpiob-2" to differentiate
	if (port > 1) {
		for (i = 1; i < port; i++) {
			if (mtac_port_info[i - 1]) {
				if (strstr(mtac_port_info[i - 1]->product_id, PRODUCT_ID_MTAC_GPIOB)) {
					count++;
				}
			}
		}
	}
	if (count > 0) {
		sprintf(buf, "gpiob-%d", count + 1);
	} else {
		sprintf(buf, "gpiob");
	}
	ret = sysfs_create_link(mtac_port_info[port_index]->subdirs->parent, mtac_port_info[port_index]->subdirs, buf);
	if (ret) {
		log_error("failed to link [%s] to [%s], %d", buf, mtac_port_info[port_index]->subdirs->name, ret);
	}

	attrs = kzalloc(sizeof(struct attribute*) * ap_gpiob_attrs_size, GFP_KERNEL);
	if (! attrs) {
		log_error("failed to allocate attribute space for port %d", port);
		return false;
	}

	// add digital inputs
	for (i = 0; i < 4; i++) {
		sprintf(buf, "din%d", i);
		attr = mtac_create_attribute(buf, MTS_ATTR_MODE_RO);
		if (! attr) {
			log_error("failed to create attribute [%s]", buf);
			return false;
		}
		attr->show = mts_attr_show_ap_din;
		attrs[index++] = &attr->attr;
	}

	// add digital outputs
	for (i = 0; i < 4; i++) {
		sprintf(buf, "dout%d", i);
		attr = mtac_create_attribute(buf, MTS_ATTR_MODE_RW);
		if (! attr) {
			log_error("failed to create attribute [%s] for GPIOB in port %d", buf, port);
			return false;
		}
		attr->show = mts_attr_show_ap_dout;
		attr->store = mts_attr_store_ap_dout;
		attrs[index++] = &attr->attr;
	}

	// add analog to digital
	for (i = 0; i < 3; i++) {
		sprintf(buf, "adc%d", i);
		attr = mtac_create_attribute(buf, MTS_ATTR_MODE_RO);
		if (! attr) {
			log_error("failed to create attribute [%s] for GPIOB in port %d", buf, port);
			return false;
		}
		attr->show = mts_attr_show_ap_adc;
		attrs[index++] = &attr->attr;
	}

	// add LEDs
	for (i = 1; i <= 2; i++) {
		sprintf(buf, "led%d", i);
		attr = mtac_create_attribute(buf, MTS_ATTR_MODE_RW);
		if (! attr) {
			log_error("failed to create attribute [%s] for GPIOB in port %d", buf, port);
			return false;
		}
		attr->show = mtac_attr_show_ap_gpio_pin;
		attr->store = mtac_attr_store_ap_gpio_pin;
		attrs[index++] = &attr->attr;
	}

	// add misc attributes
	sprintf(buf, "dout-enable");
	attr = mtac_create_attribute(buf, MTS_ATTR_MODE_RW);
	if (! attr) {
		log_error("failed to create attribute [%s] for GPIOB in port %d", buf, port);
		return false;
	}
	attr->show = mtac_attr_show_ap_gpio_pin;
	attr->store = mtac_attr_store_ap_gpio_pin;
	attrs[index++] = &attr->attr;

	sprintf(buf, "reset");
	attr = mtac_create_attribute(buf, MTS_ATTR_MODE_RW);
	if (! attr) {
		log_error("failed to create attribute [%s] for GPIOB in port %d", buf, port);
		return false;
	}
	attr->show = mtac_attr_show_ap_gpio_pin;
	attr->store = mtac_attr_store_ap_gpio_pin;
	attrs[index++] = &attr->attr;

	// add attributes for eeprom contents
	if (! mtac_add_product_info_attributes(port, attrs, &index)) {
		log_error("failed to add product info attributes for GPIOB in port %d", port);
		return false;
	}

	attrs[index] = NULL;

	mtac_port_info[port_index]->attr_group.attrs = attrs;

	// setup and register drivers
	log_debug("registering accessory card %d dout driver", port);
	sprintf(buf, "mts-io-ap%d-dout", port);
	if (! gpiob_spi_driver_setup(&gpiob_spi_drivers[port_index][dout], buf)) {
		log_error("failed to set up spi driver [%s] for GPIOB in port %d", buf, port);
		return false;
	}
	if (spi_register_driver(&gpiob_spi_drivers[port_index][dout])) {
		log_error("failed to register accessory card %d dout driver", port);
		spi_unregister_driver(&gpiob_spi_drivers[port_index][dout]);
		return false;
	}

	log_debug("registering accessory card %d din driver", port);
	sprintf(buf, "mts-io-ap%d-din", port);
	if (! gpiob_spi_driver_setup(&gpiob_spi_drivers[port_index][din], buf)) {
		log_error("failed to set up spi driver [%s] for GPIOB in port %d", buf, port);
		return false;
	}
	if (spi_register_driver(&gpiob_spi_drivers[port_index][din])) {
		log_error("failed to register accessory card %d din driver", port);
		spi_unregister_driver(&gpiob_spi_drivers[port_index][din]);
		return false;
	}

	log_debug("registering accessory card %d adc driver", port);
	sprintf(buf, "mts-io-ap%d-adc", port);
	if (! gpiob_spi_driver_setup(&gpiob_spi_drivers[port_index][adc], buf)) {
		log_error("failed to set up spi driver [%s] for GPIOB in port %d", buf, port);
		return false;
	}
	if (spi_register_driver(&gpiob_spi_drivers[port_index][adc])) {
		log_error("failed to register accessory card %d adc driver", port);
		spi_unregister_driver(&gpiob_spi_drivers[port_index][adc]);
		return false;
	}

	if (sysfs_create_group(mtac_port_info[port_index]->subdirs, &mtac_port_info[port_index]->attr_group)) {
		log_error("sysfs_create_group failed for GPIOB in port %d", port);
		return false;
	}

	return true;
}

static bool gpiob_teardown(enum ap port) {
	int i;
	int port_index = port - 1;
	struct attribute **attrs = mtac_port_info[port_index]->attr_group.attrs;

	log_info("unloading GPIOB accessory card in port %d", port);

	// clean up allocated memory for attributes
	for (i = 0; i < ap_gpiob_attrs_size; i++) {
		if (attrs[i]) {
			if (attrs[i]->name)
				kfree(attrs[i]->name);

			kfree(attrs[i]);
		}
	}

	kfree(attrs);

	// clean up our "apX/" kobject if it exists
	if (mtac_port_info[port_index]->subdirs) {
		kobject_put(mtac_port_info[port_index]->subdirs);
	}

	// clean up allocated memory for SPI drivers
	if (gpiob_spi_drivers[port_index][dout].driver.name)
		kfree(gpiob_spi_drivers[port_index][dout].driver.name);
	if (gpiob_spi_drivers[port_index][din].driver.name)
		kfree(gpiob_spi_drivers[port_index][din].driver.name);
	if (gpiob_spi_drivers[port_index][adc].driver.name)
		kfree(gpiob_spi_drivers[port_index][adc].driver.name);

	// unregister SPI drivers
	spi_unregister_driver(&gpiob_spi_drivers[port_index][dout]);
	spi_unregister_driver(&gpiob_spi_drivers[port_index][din]);
	spi_unregister_driver(&gpiob_spi_drivers[port_index][adc]);
	mtac_clear_port_pins(port_index);
	return true;
}

bool set_gpiob_info(struct ap_info* info) {
	snprintf(info->product_id, 32, "%s", PRODUCT_ID_MTAC_GPIOB);
	info->setup = &gpiob_setup;
	info->teardown = &gpiob_teardown;
	info->gpio_pin_name_by_attr_name = &gpiob_gpio_pin_name_by_attr_name;

	return true;
}

/*
 * Loop through all the slots, and set up the
 * mtac-gpiob driver for all slots.
 */
static int __init mtac_gpiob_init(void)
{
	int slot_count = 0;

	slot_count = mtac_find(set_gpiob_info,PRODUCT_ID_MTAC_GPIOB);

	if (slot_count < 1) {
		log_debug("No MTAC GPIOB found");
		if (slot_count < 0)
			return slot_count;
		else
			return -ENXIO;

	}

	return 0;
}

/* We can only tear down our own device */
static void __exit mtac_gpiob_exit(void)
{
	mtac_free(PRODUCT_ID_MTAC_GPIOB,eth_gpiob,"gpiob");
	log_info("exiting");
}

module_init(mtac_gpiob_init);
module_exit(mtac_gpiob_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_VERSION(DRIVER_VERSION);
MODULE_LICENSE("GPL");
